#!/usr/bin/python
###########################################################################################
#   textreader - read text out of various file types
#
#   Date        Author      Reason
#   ----        ------      ------
#   12/29/12    Lou King    Create
#
#   Copyright 2012 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###########################################################################################
'''
textreader - read text out of various file types
===================================================
'''

# standard
import argparse
import csv

# pypi

# github

# home grown
from . import version

# valid file types
VALIDFTYPES = ['xls','xlsx','docx','txt','csv']
# valid list types
VALIDLTYPES = ['txt','csv']

DOCXTABSIZE = 8
TXTABSIZE = 8

# exceptions for this module.  See __init__.py for package exceptions
class headerError(Exception): pass
class parameterError(Exception): pass

########################################################################
class TextDictReader():
########################################################################
    '''
    get data from a file in various formats
    
    :param filename: name of file to open, or list-like
    :param fieldxform: {'fieldname1':['hdr11, hdr12, ...'], 'fieldname2':[hdr21, ...], ...}, where fieldnamex will be returned in dict, hdrxx are possible headers for columns
    :param reqdfields: list of fields required to be in the header
    :param filetype: if filename is list, this should be filetype list should be interpreted as
    '''
    #----------------------------------------------------------------------
    def __init__(self, filename, fieldxform, reqdfields, filetype='extension'):
    #----------------------------------------------------------------------
        # open the textreader using the file
        self.file = TextReader(filename, filetype)
        self.fieldxform = fieldxform
        self.reqdfields = reqdfields
        
        # self.field item value will be of form {'begin':startindex,'end':startindex+length} for easy slicing
        self.field = {}

        # scan to the header line
        self._findhdr()

    #----------------------------------------------------------------------
    def _findhdr(self):
    #----------------------------------------------------------------------
        '''
        find the header in the file
        '''
    
        foundhdr = False
        delimited = self.file.getdelimited()
        fields = list(self.fieldxform.keys())
        MINMATCHES = len(self.reqdfields)

        # catch StopIteration, which means header wasn't found in the file
        try:
            # loop for each line until header found
            while True:
                origline = next(self.file)
                fieldsfound = 0
                line = []
                if not delimited:
                    for word in origline.split():
                        line.append(word.lower())
                else:
                    for word in origline:
                        line.append(str(word).lower())  # str() called in case non-string returned in origline
                    
                # loop for each potential self.field in a header
                for fieldndx in range(len(fields)):
                    f = fields[fieldndx]
                    match = self.fieldxform[f]
                    
                    # loop for each match possiblity, then within the fields in file header
                    # folding loop likes this gives precedence to the list order of the match possibilities
                    for m in match:
                        # loop for each column in this line, trying to find match possiblities
                        matchfound = False
                        for linendx in range(len(line)):                        
                            # match over the end of the line is no match
                            # m is either a string or a list of strings
                            if isinstance(m, str):
                                m = [m]         # make single string into list
                            if linendx+len(m)>len(line):
                                continue
                            # if we found the match, remember start and end of match
                            if line[linendx:linendx+len(m)] == m:
                                if f not in self.field: self.field[f] = {}
                                self.field[f]['start'] = linendx
                                self.field[f]['end'] = linendx + len(m)
                                self.field[f]['match'] = m
                                self.field[f]['genfield'] = f   # seems redundant, but [f] index is lost later in self.foundfields
                                fieldsfound += 1
                                matchfound = True
                                break   # match possibility loop
                        
                        # found match for this self.field
                        if matchfound: break
                
                # here we've gone through each self.field in the line
                # need to match more than MINMATCHES to call it a header line
                if fieldsfound >= MINMATCHES:
                    
                    # verify that all other required fields are present
                    fieldsnotfound = []
                    for f in self.reqdfields:
                        if f not in self.field:
                            fieldsnotfound.append(f)
                    if len(fieldsnotfound) != 0:
                        raise headerError('could not find fields {} in header {}'.format(fieldsnotfound, origline))
                        
                    # sort found fields by order found within the line
                    foundfields_dec = sorted([(self.field[f]['start'],self.field[f]) for f in self.field])
                    self.foundfields = [ff[1] for ff in foundfields_dec] # get rid of sorting decorator
                        
                    # here we have decided it is a header line
                    # if the file is not delimited, we have to find where these fields start
                    # and tell self.file where the self.field breaks are
                    # assume multi self.field matches are separated by single space
                    if not delimited:
                        # sort found fields by 'start' linendx (self.field number within line)
                        # loop through characters in original line, skipping over spaces within matched fields, to determine
                        # where delimiters should be
                        delimiters = []
                        thischar = 0
                        foundfields_iter = iter(self.foundfields)
                        thisfield = next(foundfields_iter)
                        while True:
                            # scan past the white space
                            while thischar < len(origline) and origline[thischar] == ' ': thischar += 1
                            
                            # we're done looking if we're at the end of the line
                            if thischar == len(origline): break
                            
                            # found a word, remember where it was
                            delimiters.append(thischar)
                            
                            # look for the next match of known header fields
                            matchfound = False
                            if thisfield is not None:
                                # if a match, might be multiple words.  Probably ok to assume single space between them
                                fullmatch = ' '.join(thisfield['match'])
                                if origline[thischar:thischar+len(fullmatch)].lower() == fullmatch:
                                    thischar += len(fullmatch)
                                    matchfound = True
                                    try:
                                        thisfield = next(foundfields_iter)
                                    except StopIteration:
                                        thisfield = None
                            
                            # if found a match, thischar is already updated.  Otherwise, scan past this word
                            if not matchfound:
                                while thischar < len(origline) and origline[thischar] != ' ': thischar += 1
                            
                            # we're done looking if we're at the end of the line
                            if thischar == len(origline): break
                        
                        # set up delimiters in the file reader
                        self.file.setdelimiter(delimiters)
                                    
                    break

            # header fields are in foundfields
            # need to figure out the indeces for data which correspond to the foundfields
            self.fieldhdrs = []
            self.fieldcols = []
            skipped = 0
            for f in self.foundfields:
                self.fieldhdrs.append(f['genfield'])
                currcol = f['start'] - skipped
                self.fieldcols.append(currcol)
                skipped += len(f['match']) - 1  # if matched multiple columns, need to skip some
                
        # not good to come here
        except StopIteration:
            raise headerError('header not found')
        
    #----------------------------------------------------------------------
    def __next__(self):
    #----------------------------------------------------------------------
        '''
        return dict with generic headers and associated data from file
        '''
        
        # get next raw line from the file
        rawline = next(self.file)
        
        # pick columns which are associated with generic headers
        filteredline = [rawline[i] for i in range(len(rawline)) if i in self.fieldcols]
        
        # create dict association, similar to csv.DictReader
        result = dict(list(zip(self.fieldhdrs,filteredline)))
                   
        # and return result
        return result

########################################################################
class TextReader():
########################################################################
    '''
    abstract text reader for several different file types
    creation of object opens TextReader file
    
    :param filename: name of file to open, or list
    '''

    #----------------------------------------------------------------------
    def __init__(self, filename, filetype='extension'):
    #----------------------------------------------------------------------
        """
        open TextReader file
        
        :param filename: name of file to open, or list-like
        :param filetype: if filename is list, this should be filetype list should be interpreted as
        """

        # if true filename, type of filename is string
        if isinstance(filename, str):
            self.ftype = filename.split('.')[-1].lower() # get extension
            self.intype = 'file'
            if self.ftype not in VALIDFTYPES:
                raise parameterError('Invalid filename {}: must have extension in {}'.format(filename, VALIDFTYPES))
        # otherwise assume 'filename' is list-like
        else:
            self.ftype = filetype.lower()
            self.intype = 'list'
            if self.ftype not in VALIDLTYPES:
                raise parameterError('Invalid list: must use filetype in {}'.format(VALIDLTYPES))
        
        # handle excel files
        if self.ftype in ['xls']:
            from xlrd import open_workbook
            self.workbook = open_workbook(filename)
            self.sheet = self.workbook.sheet_by_index(0)    # only first sheet is considered
            self.currrow = 0
            self.nrows = self.sheet.nrows
            self.delimited = True               # rows are already broken into columns
            self.workbook.release_resources()   # sheet is already loaded so we can save memory
            
        # handle excel files
        elif self.ftype in ['xlsx']:
            from openpyxl import load_workbook
            self.workbook = load_workbook(filename)
            self.sheet = self.workbook[self.workbook.sheetnames[0]]    # only first sheet is considered
            self.currrow = 0
            self.rows = list(self.sheet.rows)
            self.nrows = len(list(self.sheet.rows))
            self.delimited = True               # rows are already broken into columns

        # handle word files
        elif self.ftype in ['docx']:
            import docx
            doc = docx.opendocx(filename)
            self.lines = iter(docx.getdocumenttext(doc))
            self.delimited = False
            
        # handle txt files
        elif self.ftype in ['txt']:
            if self.intype == 'file':
                self.TXT = open(filename,'r')
            else:
                self.TXT = iter(filename)
            self.delimited = False
            
        # handle txt files
        elif self.ftype in ['csv']:
            if self.intype == 'file':
                self._CSV = open(filename, 'r', encoding='utf8')
            else:
                self._CSV = iter(filename)
            self.CSV = csv.reader(self._CSV)
            self.delimited = True
            
        self.delimiters = None
        self.opened = True
        
    #----------------------------------------------------------------------
    def __next__(self):
    #----------------------------------------------------------------------
        """
        read next line from TextReader file
        raises StopIteration when end of file reached
        
        :rtype: list of cells for the current row
        """
        
        # check if it is ok to read
        if not self.opened:
            raise ValueError('I/O operation on a closed file')
        
        # handle excel files
        if self.ftype in ['xls']:
            if self.currrow >= self.nrows:
                raise StopIteration
            
            row = self.sheet.row_values(self.currrow)
            self.currrow += 1
            return row
        
        # handle excel files
        if self.ftype in ['xlsx']:
            if self.currrow >= self.nrows:
                raise StopIteration
            
            row = [c.value for c in self.rows[self.currrow]]
            self.currrow += 1
            return row
        
        # handle word files
        elif self.ftype in ['docx']:
            line = next(self.lines)
            line = line.expandtabs(DOCXTABSIZE)
            if self.delimiters:
                splitline = self.delimit(line)
                return splitline
            else:
                return line

        # handle txt files
        elif self.ftype in ['txt']:
            line = next(self.TXT)       # TODO: can this be self.TXT.next() ?
            line = line.expandtabs(TXTABSIZE)
            if self.delimiters:
                splitline = self.delimit(line)
                return splitline
            else:
                return line

        # handle txt files
        elif self.ftype in ['csv']:
            row = next(self.CSV)
            return row

    #----------------------------------------------------------------------
    def close(self):
    #----------------------------------------------------------------------
        """
        close TextReader file
        """
        
        self.opened = False
        
        # handle lists
        if self.intype == 'list':
            pass    # no resources to release

        # handle excel files
        elif self.ftype in ['xls','xlsx']:
            pass    # resources have already been released
        
        # handle word files
        elif self.ftype in ['docx']:
            pass    # nothing going on here, either
        
        # handle txt files
        elif self.ftype in ['txt']:
            self.TXT.close()
        
        # handle txt files
        elif self.ftype in ['csv']:
            self._CSV.close()
        

    #----------------------------------------------------------------------
    def getdelimited(self):
    #----------------------------------------------------------------------
        """
        return state whether file is delimited
        
        :rtype: True if delimited
        """
        
        return self.delimited       

    #----------------------------------------------------------------------
    def setdelimiter(self,delimiters):
    #----------------------------------------------------------------------
        """
        set delimiters for file as specified in delimiters
        
        :param delimiters: list of character positions to set delimiters at
        """
        
        # do some validation
        if self.delimited:
            raise parameterError('cannot set delimiters for file which is already delimited')
        if not isinstance(delimiters, list):
            raise parameterError('delimiters must be a list of increasing integers')
        lastdelimiter = -1
        for delimiter in delimiters:
            if not isinstance(delimiter, int) or delimiter <= lastdelimiter:
                raise parameterError('delimiters must be a list of increasing integers')
            lastdelimiter = delimiter
            
        self.delimiters = delimiters
        self.delimited = True
        
    #----------------------------------------------------------------------
    def delimit(self,s):
    #----------------------------------------------------------------------
        """
        split a string based on delimiters
        
        :param s: string to be split
        :rtype: list of string elements, stripped of white space
        """
        
        if not self.delimiters:
            raise parameterError('cannot split string if delimiters not set')
        
        rval = []
        for i in range(len(self.delimiters)):
            start = self.delimiters[i]
            end = self.delimiters[i+1] if i+1 < len(self.delimiters) else None   # last one goes to end of s
            rval.append(s[start:end].strip())
            
        return rval
            
################################################################################
def main():
################################################################################

    parser = argparse.ArgumentParser(version='{0} {1}'.format('loutilities',version.__version__))
    parser.add_argument('filename',help='name of file for testing')
    args = parser.parse_args()
    
    # act on arguments
    filename = args.filename
    
    # open file, print some lines, then close
    ff = TextReader(filename)
    for i in range(6):
        print(next(ff))
    ff.close()

# ###############################################################################
# ###############################################################################
if __name__ == "__main__":
    main()




###########################################################################################
#   configparser - enhanced ConfigParser
#
#   Date        Author      Reason
#   ----        ------      ------
#   06/20/16    Lou King    create
#
#   Copyright 2016 Lou King
###########################################################################################

'''
configparser - enhanced ConfigParser
===================================================
uses pypi's configparser.ConfigParser.SafeConfigParser to parse INI file, but preserves case for keys
'''

# standard
from collections import OrderedDict
from configparser import ConfigParser

# pypi

config = ConfigParser()

# preserve case for keys, see https://docs.python.org/3/library/configparser.html#configparser.optionxform
config.optionxform = lambda option: option

# copy configuration into dict
def getitems(filepath, section):
    '''
    get items in section
    convert to integer, float, etc., as appropriate
    
    :param filepath: file to read config from
    :param section: section to read items from
    :rtype: OrderedDict containing {key:item, ...} (case is preserved for keys)
    '''
    config.readfp(open(filepath))
    thisconfig = config.items(section)
    outdict = OrderedDict()

    # apply configuration to app
    # eval is safe because this configuration is controlled at root
    for key,value in thisconfig:
        try:
            outdict[key] = eval(value)
        except:
            outdict[key] = value

    return outdict




'''
user - package supports user management for xtilities products
=========================================================================================
'''

# pypi
from flask import Flask, g
from flask_security import Security, SQLAlchemyUserDatastore, LoginForm, ForgotPasswordForm

# homegrown
from loutilities.configparser import getitems
from loutilities.user.model import User, Role

# hold application here
app = None
user_datastore = None

# TODO: should these messages be localized? See https://flask-security-too.readthedocs.io/en/stable/customizing.html#localization
user_messages = {
    'ACCOUNT_NOT_PERMITTED' : 'Account not permitted for this application'
}

# login_form for application management
class UserLoginForm(LoginForm):
    def validate(self):
        # if some error was detected from standard validate(), we're done
        if not super().validate():
            return False

        # if all ok otherwise, check roles to verify user allowed for this application
        ## collect applications
        apps = set()
        for thisrole in self.user.roles:
            apps |= set(thisrole.applications)
        ## disallow login if this app isn't in one of user's roles
        if g.loutility not in apps:
            self.email.errors.append(user_messages['ACCOUNT_NOT_PERMITTED'])
            return False

        return True

# forgot_password for application management
class UserForgotPasswordForm(ForgotPasswordForm):
    def validate(self):
        # if some error was detected from standard validate(), we're done
        if not super().validate():
            return False

        # if all ok otherwise, check roles to verify user allowed for this application
        ## collect applications
        apps = set()
        for thisrole in self.user.roles:
            apps |= set(thisrole.applications)
        ## disallow login if this app isn't in one of user's roles
        if g.loutility not in apps:
            self.email.errors.append(user_messages['ACCOUNT_NOT_PERMITTED'])
            return False

        return True

# extend flask_security.Security to support application verification
class UserSecurity(Security):
    def __init__(self, app=None, datastore=None, register_blueprint=True, **kwargs):
        '''
        replaces flask_security.Security

        add login_form=UserLoginForm if caller hasn't already supplied
        :param kwargs:
        '''
        if not 'login_form' in kwargs:
            kwargs['login_form'] = UserLoginForm
        if not 'forgot_password_form' in kwargs:
            kwargs['forgot_password_form'] = UserForgotPasswordForm
        return super().__init__(app, datastore, register_blueprint, **kwargs)

# used only for database initialization
# TODO: future use for loutilities.com landing page
def create_app(config_obj, configfiles=None):
    '''
    apply configuration object, then configuration files
    '''
    global app
    app = Flask('loutilities')
    app.config.from_object(config_obj)
    if configfiles:
        # backwards compatibility
        if type(configfiles) == str:
            configfiles = [configfiles]
        for configfile in configfiles:
            appconfig = getitems(configfile, 'app')
            app.config.update(appconfig)

    from .model import db
    db.init_app(app)

    global user_datastore
    user_datastore = SQLAlchemyUserDatastore(db, User, Role)
    security = Security(app, user_datastore)

    # need to force app context else get
    #    RuntimeError: Working outside of application context.
    #    RuntimeError: Attempted to generate a URL without the application context being pushed.
    # see http://kronosapiens.github.io/blog/2014/08/14/understanding-contexts-in-flask.html
    with app.app_context():
        # set up scoped session
        from sqlalchemy.orm import scoped_session, sessionmaker
        # the following code causes binds not to work, because the session is artificially
        # set to the base database engine via bind parameter
        # db.session = scoped_session(sessionmaker(autocommit=False,
        #                                          autoflush=False,
        #                                          bind=db.engine))
        # db.query = db.session.query_property()

    return app


###########################################################################################
# roles - common location for xtility role declaration
#
#       Date            Author          Reason
#       ----            ------          ------
#       03/11/20        Lou King        Create
#
#   Copyright 2020 Lou King.  All rights reserved
###########################################################################################

from loutilities.user.model import APP_CONTRACTS, APP_MEMBERS, APP_ROUTES, APP_SCORES, APP_ALL

# common roles
ROLE_SUPER_ADMIN = 'super-admin'
ROLES_COMMON = [ROLE_SUPER_ADMIN]
roles_common = [
    {'name': 'super-admin', 'description': 'allowed to do everything on all applications', 'apps': APP_ALL},
]

# members roles
ROLE_LEADERSHIP_ADMIN = 'leadership-admin'
ROLE_LEADERSHIP_MEMBER = 'leadership-member'
ROLE_MEMBERSHIP_ADMIN = 'membership-admin'
ROLE_MEETINGS_ADMIN = 'meetings-admin'
ROLE_MEETINGS_MEMBER = 'meetings-member'
roles_members = [
    {'name': ROLE_LEADERSHIP_ADMIN, 'description': 'access to leadership tasks for members application', 'apps':[APP_MEMBERS]},
    {'name': ROLE_LEADERSHIP_MEMBER, 'description': 'user of leadership tasks for members application', 'apps':[APP_MEMBERS]},
    {'name': ROLE_MEMBERSHIP_ADMIN, 'description': 'access to membership admininstration for members application', 'apps':[APP_MEMBERS]},
    {'name': ROLE_MEETINGS_ADMIN, 'description': 'access to meetings administration for members application', 'apps':[APP_MEMBERS]},
    {'name': ROLE_MEETINGS_MEMBER, 'description': 'user of meetings for members application', 'apps':[APP_MEMBERS]},
]

# routes roles
ROLE_ROUTES_ADMIN = 'routes-admin'
ROLE_ICON_ADMIN = 'icon-admin'
roles_routes = [{'name': ROLE_ROUTES_ADMIN, 'description': 'access to routes for routes application', 'apps':[APP_ROUTES]},
                 {'name': ROLE_ICON_ADMIN, 'description': 'access to icons for routes application', 'apps':[APP_ROUTES]}
                 ]

# contracts roles
ROLE_EVENT_ADMIN = 'event-admin'
ROLE_SPONSOR_ADMIN = 'sponsor-admin'
roles_contracts = [{'name': ROLE_EVENT_ADMIN, 'description': 'access to events for contracts application', 'apps':[APP_CONTRACTS]},
                   {'name': ROLE_SPONSOR_ADMIN, 'description': 'access to sponsors/races for contracts application', 'apps':[APP_CONTRACTS]}
                   ]

# scores roles
ROLE_SCORES_ADMIN = 'scores-admin'
ROLE_SCORES_VIEWER = 'scores-viewer'
roles_scores = [{'name': ROLE_SCORES_ADMIN, 'description': 'administer scores application', 'apps':[APP_SCORES]},
                   {'name': ROLE_SCORES_VIEWER, 'description': 'view scores application', 'apps':[APP_SCORES]},
                   ]

all_roles = [roles_common, roles_contracts, roles_members, roles_routes, roles_scores]


###########################################################################################
# userrole - manage application users and roles
#
#       Date            Author          Reason
#       ----            ------          ------
#       12/08/19        Lou King        Create
#
#   Copyright 2019 Lou King
#
###########################################################################################
'''
userrole - manage application users and roles
====================================================
'''

# standard

# pypi
from validators.slug import slug
from validators.email import email
from flask import g, request
from flask_security import current_user
from flask_security.recoverable import send_reset_password_instructions

# homegrown
from . import bp
from loutilities.user.model import db, User, Role, Interest, Application
from loutilities.tables import DbCrudApiRolePermissions, SEPARATOR, DteDbRelationship
from loutilities.timeu import asctime
from ..roles import ROLE_SUPER_ADMIN

ymdtime = asctime('%Y-%m-%d %H:%M:%S')

class ParameterError(Exception):
    '''raise exception if parameter error'''

##########################################################################################
# users endpoint
###########################################################################################

user_dbattrs = 'id,email,name,given_name,roles,interests,last_login_at,current_login_at,last_login_ip,current_login_ip,login_count,active'.split(',')
user_formfields = 'rowid,email,name,given_name,roles,interests,last_login_at,current_login_at,last_login_ip,current_login_ip,login_count,active'.split(',')
user_dbmapping = dict(zip(user_dbattrs, user_formfields))
user_formmapping = dict(zip(user_formfields, user_dbattrs))

user_formmapping['last_login_at'] = lambda dbrow: ymdtime.dt2asc(dbrow.last_login_at) if dbrow.last_login_at else ''
user_formmapping['current_login_at'] = lambda dbrow: ymdtime.dt2asc(dbrow.current_login_at) if dbrow.current_login_at else ''

def user_validate(action, formdata):
    results = []

    if formdata['email'] and not email(formdata['email']):
        results.append({ 'name' : 'email', 'status' : 'invalid email: correct format is like john.doe@example.com' })

    # check apps which user will have access to
    apps = set()
    if formdata['roles'] and 'id' in formdata['roles'] and formdata['roles']['id'] != '':
        roleidsstring = formdata['roles']['id']
        roleids = roleidsstring.split(SEPARATOR)
        for roleid in roleids:
            thisrole = Role.query.filter_by(id=roleid).one()
            apps |= set(thisrole.applications)

    # this app must be one of user's roles
    if g.loutility not in apps:
        # need to use name='roles.id' because this field is _treatment:{relationship}
        results.append({'name': 'roles.id', 'status': 'give user at least one role which works for this application'})

    return results

class RolesPicker(DteDbRelationship):
    '''
    pick Roles, but special processing based on ROLE_SUPER_ADMIN, i.e., if not ROLE_SUPER_ADMIN only present
    roles allowed for this application
    '''

    def __init__(self, **kwargs):
        # the args dict has default values for arguments added by this derived class
        # caller supplied keyword args are used to update these
        # all arguments are made into attributes for self by the inherited class
        args = dict(
            tablemodel=User,
            fieldmodel=Role,
            labelfield='name',
            formfield='roles',
            dbfield='roles',
            uselist=True,
        )
        args.update(kwargs)

        # this initialization needs to be done before checking any self.xxx attributes
        super().__init__(**args)

    def allowed_roles(self):
        # create a copy so we're not messing with Application record, no more can be configured than current users'
        allowed_roles = current_user.roles[:]
        return allowed_roles

    def set(self, formrow):
        '''
        if not ROLE_SUPER_ADMIN merge newly set roles with those user can't see
        '''
        # these are the roles from the form, but limited to allowed_roles if not ROLE_SUPER_ADMIN
        resultroles = super().set(formrow)
        if not current_user.has_role(ROLE_SUPER_ADMIN):
            theuser = User.query.filter_by(email=formrow['email']).one_or_none()
            allowed_roles = self.allowed_roles()
            if theuser:
                otherroles = [r for r in theuser.roles if r not in allowed_roles]
                resultroles += otherroles
        return resultroles

    def get(self, dbrow_or_id):
        '''
        if not ROLE_SUPER_ADMIN only return roles allowed for this user
        '''
        rv = super().get(dbrow_or_id)
        rvnames = rv['name'].split(SEPARATOR)
        rvids = rv['id'].split(SEPARATOR)
        if not current_user.has_role(ROLE_SUPER_ADMIN):
            allowed_roles = self.allowed_roles()
            allowed_role_names = [r.name for r in allowed_roles]
            allowed_role_ids = [str(r.id) for r in allowed_roles]
            rv = {
                'name': SEPARATOR.join([item for item in rvnames if item in allowed_role_names]),
                'id': SEPARATOR.join([item for item in rvids if item in allowed_role_ids])
            }
        return rv

    def options(self):
        '''limit visible options to what user can see if not ROLE_SUPER_ADMIN'''
        opts = super().options()
        if not current_user.has_role(ROLE_SUPER_ADMIN):
            allowed_roles = self.allowed_roles()
            allowed_role_ids = [r.id for r in allowed_roles]
            opts = [o for o in opts if o['value'] in allowed_role_ids]
        return opts


class UserCrudApi(DbCrudApiRolePermissions):
    '''
    extends DbCrudApiRolePermissions to manage user for single sign-on with flask-security

    Additional parameters for this class:

    :param user_datastore: return value from flask_security.SQLAlchemyUserDatastore(db, User, Role), which needs
        to be called from application using single sign-on
    '''

    def __init__(self, **kwargs):
        # the args dict has default values for arguments added by this derived class
        # caller supplied keyword args are used to update these
        # all arguments are made into attributes for self by the inherited class
        args = dict(
                    user_datastore=None,
                    )
        args.update(kwargs)

        # this initialization needs to be done before checking any self.xxx attributes
        super().__init__(**args)

        # Caller must use local_interest_model
        if not self.user_datastore:
            raise ParameterError('user_datastore required')

    def createrow(self, formdata):
        '''
        createrow is used by create form, may need to also send password reset request to user.
        comes from tables-assets/static/user/admin/beforedatatables.js user_create_send_notification_button()

        :param formdata: data from form
        :return:
        '''
        # create the user
        email = formdata['email']
        name = formdata['name']
        given_name = formdata['given_name']
        roles = [Role.query.filter_by(id=id).one() for id in formdata['roles']['id'].split(SEPARATOR)]
        # ROLE_SUPER_ADMIN can set interests
        if current_user.has_role(ROLE_SUPER_ADMIN):
            interests = [Interest.query.filter_by(id=id).one() for id in formdata['interests']['id'].split(SEPARATOR)]
        # non ROLE_SUPER_ADMIN uses currently selected interest for user (interests is not on form)
        else:
            interests = [Interest.query.filter_by(interest=g.interest).one()]
        newuser = self.user_datastore.create_user(email=email, name=name, given_name=given_name,
                                                  roles=roles, interests=interests)

        # force id to be set. required for response data and password reset
        db.session.flush()

        # return the newly created row
        row = self.dte.get_response_data(newuser)

        # admin may have requested password reset email be sent to the user
        if 'resetpw' in request.form:
            send_reset_password_instructions(newuser)

        return row

    def check_superadmin(self, col):
        '''
        check if col should be included in display based on user's roles

        :param col: column to check
        :return: True if column should be included
        '''
        rv = True
        if not current_user.has_role(ROLE_SUPER_ADMIN):
            supercols = ['interests', 'last_login_at', 'last_login_ip', 'current_login_ip', 'login_count']
            colname = col['name'].split('.')[0]
            if colname in supercols:
                rv = False
        return rv

    def getdtoptions(self):
        '''limit columns to those this user is allowed to see'''
        dtoptions = super().getdtoptions()
        dtoptions['columns'] = [c for c in dtoptions['columns'] if self.check_superadmin(c)]
        return dtoptions

    def getedoptions(self):
        '''limit form fields to those this user is allowed to see'''
        edoptions = super().getedoptions()
        edoptions['fields'] = [c for c in edoptions['fields'] if self.check_superadmin(c)]
        return edoptions

    def updaterow(self, thisid, formdata):
        '''
        updaterow is used by edit form, may need to also send password reset request to user.
        comes from tables-assets/static/user/admin/beforedatatables.js reset_password_button()

        :param thisid: id of user
        :param formdata: edit form
        :return: row data
        '''
        if 'resetpw' in request.form:
            user = User.query.filter_by(id=thisid).one()
            send_reset_password_instructions(user)
        return super().updaterow(thisid, formdata)

class UserView(UserCrudApi):
    def __init__(self, **kwargs):
        '''
        application MUST instantiate UserView

        application should override editor_method_postcommit(self, form) to call
        loutilities.model.ManageLocalUser(db, appname, localusermodel, localinterestmodel).update()
        '''
        self.kwargs = kwargs
        args = dict(
            app=bp,  # use blueprint instead of app
            db=db,
            model=User,
            version_id_col='version_id',  # optimistic concurrency control
            roles_accepted=ROLE_SUPER_ADMIN,
            template='datatables.jinja2',
            pagename='users',
            endpoint='userrole.users',
            rule='/users',
            dbmapping=user_dbmapping,
            formmapping=user_formmapping,
            clientcolumns=[
                {'data': 'email', 'name': 'email', 'label': 'Email', '_unique': True,
                 'className': 'field_req',
                 },
                {'data': 'given_name', 'name': 'given_name', 'label': 'First Name',
                 'className': 'field_req',
                 },
                {'data': 'name', 'name': 'name', 'label': 'Full Name',
                 'className': 'field_req',
                 },
                {'data': 'roles', 'name': 'roles', 'label': 'Roles',
                 '_treatment': {'relationship': {'optionspicker': RolesPicker()}}
                 },
                {'data': 'interests', 'name': 'interests', 'label': 'Interests',
                 '_treatment': {'relationship': {'fieldmodel': Interest, 'labelfield': 'description',
                                                 'formfield': 'interests', 'dbfield': 'interests',
                                                 'uselist': True}}
                 },
                {'data': 'active', 'name': 'active', 'label': 'Active',
                 '_treatment': {'boolean': {'formfield': 'active', 'dbfield': 'active'}},
                 'ed': {'def': 'yes'},
                 },
                {'data': 'last_login_at', 'name': 'last_login_at', 'label': 'Last Login At',
                 'className': 'dt-body-nowrap',
                 'type': 'readonly'
                 },
                {'data': 'current_login_at', 'name': 'current_login_at', 'label': 'Current Login At',
                 'className': 'dt-body-nowrap',
                 'type': 'readonly',
                 },
                {'data': 'last_login_ip', 'name': 'last_login_ip', 'label': 'Last Login IP', 'type': 'readonly'},
                {'data': 'current_login_ip', 'name': 'current_login_ip', 'label': 'Current Login IP',
                 'type': 'readonly'
                 },
                {'data': 'login_count', 'name': 'login_count', 'label': 'Login Count', 'type': 'readonly'},
            ],
            validate=user_validate,
            servercolumns=None,  # not server side
            idSrc='rowid',
            buttons=[{
                         'extend': 'create',
                         'editor': {'eval': 'editor'},
                         'formButtons': [
                             {'text': 'Create and Send', 'action': {'eval': 'user_create_send_notification_button'}},
                             {'text': 'Create', 'action': {'eval': 'submit_button'}},
                         ]
                     },
                     {
                         'extend': 'editRefresh',
                         'text': 'Edit',
                         'editor': {'eval': 'editor'},
                         'formButtons': [
                             {'text': 'Reset Password', 'action': {'eval': 'reset_password_button'}},
                             {'text': 'Update', 'action': {'eval': 'submit_button'}},
                         ]
                     },
            ],
            dtoptions={
                'scrollCollapse': True,
                'scrollX': True,
                'scrollXInner': "100%",
                'scrollY': True,
            },
        )
        args.update(kwargs)
        super().__init__(**args)


##########################################################################################
# roles endpoint
###########################################################################################

role_dbattrs = 'id,name,description,applications'.split(',')
role_formfields = 'rowid,name,description,applications'.split(',')
role_dbmapping = dict(zip(role_dbattrs, role_formfields))
role_formmapping = dict(zip(role_formfields, role_dbattrs))

class RoleView(DbCrudApiRolePermissions):
    def __init__(self, **kwargs):
        '''
        application MUST instantiate RoleView

        application should override editor_method_postcommit(self, form) to call
        loutilities.model.ManageLocalUser(db, appname, localusermodel, localinterestmodel).update()
        '''
        self.kwargs = kwargs
        args = dict(
                    app = bp,   # use blueprint instead of app
                    db = db,
                    model = Role, 
                    version_id_col = 'version_id',  # optimistic concurrency control
                    roles_accepted = ROLE_SUPER_ADMIN,
                    template = 'datatables.jinja2',
                    pagename = 'roles', 
                    endpoint = 'userrole.roles',
                    rule = '/roles',
                    dbmapping = role_dbmapping, 
                    formmapping = role_formmapping, 
                    clientcolumns = [
                        { 'data': 'name', 'name': 'name', 'label': 'Name',
                          'className': 'field_req',
                          },
                        { 'data': 'description', 'name': 'description', 'label': 'Description' },
                        {'data': 'applications', 'name': 'applications', 'label': 'Applications',
                         '_treatment': {'relationship': {'fieldmodel': Application, 'labelfield': 'application',
                                                         'formfield': 'applications', 'dbfield': 'applications',
                                                         'uselist': True}}
                         },
                    ],
                    servercolumns = None,  # not server side
                    idSrc = 'rowid', 
                    buttons = ['create', 'editRefresh', 'remove'],
                    dtoptions = {
                                        'scrollCollapse': True,
                                        'scrollX': True,
                                        'scrollXInner': "100%",
                                        'scrollY': True,
                                  },
                    )
        args.update(kwargs)
        super().__init__(**args)

##########################################################################################
# interests endpoint
###########################################################################################

interest_dbattrs = 'id,interest,description,users,public,applications'.split(',')
interest_formfields = 'rowid,interest,description,users,public,applications'.split(',')
interest_dbmapping = dict(zip(interest_dbattrs, interest_formfields))
interest_formmapping = dict(zip(interest_formfields, interest_dbattrs))

def interest_validate(action, formdata):
    results = []

    for field in ['interest']:
        if formdata[field] and not slug(formdata[field]):
            results.append({ 'name' : field, 'status' : 'invalid slug: must be only alpha, numeral, hyphen' })

    return results

class InterestView(DbCrudApiRolePermissions):
    def __init__(self, **kwargs):
        '''
        application MUST instantiate InterestView

        application should override editor_method_postcommit(self, form) to call
        loutilities.model.ManageLocalUser(db, appname, localusermodel, localinterestmodel).update()
        '''
        self.kwargs = kwargs
        args = dict(
            app=bp,  # use blueprint instead of app
            db=db,
            model=Interest,
            version_id_col='version_id',  # optimistic concurrency control
            interests_accepted=ROLE_SUPER_ADMIN,
            template='datatables.jinja2',
            pagename='interests',
            endpoint='userrole.interests',
            rule='/interests',
            dbmapping=interest_dbmapping,
            formmapping=interest_formmapping,
            clientcolumns=[
                {'data': 'description', 'name': 'description', 'label': 'Description', '_unique': True,
                 'className': 'field_req',
                 },
                {'data': 'interest', 'name': 'interest', 'label': 'Slug', '_unique': True,
                 'className': 'field_req',
                 },
                {'data': 'public', 'name': 'public', 'label': 'Public',
                 '_treatment': {'boolean': {'formfield': 'public', 'dbfield': 'public'}},
                 'ed': {'def': 'yes'},
                 },
                {'data': 'applications', 'name': 'applications', 'label': 'Applications',
                 '_treatment': {'relationship': {'fieldmodel': Application, 'labelfield': 'application',
                                                 'formfield': 'applications', 'dbfield': 'applications',
                                                 'uselist': True}}
                 },
                {'data': 'users', 'name': 'users', 'label': 'Users',
                 '_treatment': {'relationship': {'fieldmodel': User, 'labelfield': 'email',
                                                 'formfield': 'users', 'dbfield': 'users',
                                                 'uselist': True}}
                 },
            ],
            validate=interest_validate,
            servercolumns=None,  # not server side
            idSrc='rowid',
            buttons=['create', 'editRefresh', 'remove'],
            dtoptions={
                'scrollCollapse': True,
                'scrollX': True,
                'scrollXInner': "100%",
                'scrollY': True,
            },
        )
        args.update(kwargs)
        super().__init__(**args)

##########################################################################################
# applications endpoint
###########################################################################################

application_dbattrs = 'id,application'.split(',')
application_formfields = 'rowid,application'.split(',')
application_dbmapping = dict(zip(application_dbattrs, application_formfields))
application_formmapping = dict(zip(application_formfields, application_dbattrs))

def application_validate(action, formdata):
    results = []

    for field in ['application']:
        if formdata[field] and not slug(formdata[field]):
            results.append({ 'name' : field, 'status' : 'invalid slug: must be only alpha, numeral, hyphen' })

    return results

application_view = DbCrudApiRolePermissions(
                    app = bp,   # use blueprint instead of app
                    db = db,
                    model = Application,
                    version_id_col = 'version_id',  # optimistic concurrency control
                    applications_accepted = ROLE_SUPER_ADMIN,
                    template = 'datatables.jinja2',
                    pagename = 'applications', 
                    endpoint = 'userrole.applications',
                    rule = '/applications',
                    dbmapping = application_dbmapping, 
                    formmapping = application_formmapping, 
                    clientcolumns = [
                        { 'data': 'application', 'name': 'application', 'label': 'Application', '_unique': True,
                          'className': 'field_req',
                          },
                    ],
                    validate = application_validate,
                    servercolumns = None,  # not server side
                    idSrc = 'rowid', 
                    buttons = ['create', 'editRefresh', 'remove'],
                    dtoptions = {
                                        'scrollCollapse': True,
                                        'scrollX': True,
                                        'scrollXInner': "100%",
                                        'scrollY': True,
                                  },
                    )
application_view.register()




#!/usr/bin/python
###########################################################################################
#   applytemplate - apply a template to files in a directory
#
#   Date        Author      Reason
#   ----        ------      ------
#   04/21/13    Lou King    Create
#
#   Copyright 2013 Lou King
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###########################################################################################
'''
applytemplate - apply a template to files in a directory
===============================================================

Usage::

    applytemplate [-h] [-v] [-e EXTENSION]
    
    optional arguments:
      -h, --help            show this help message and exit
      -v, --version         show program's version number and exit
      -e EXTENSION, --extension EXTENSION
                            extension for files to be processed (default htmli)
                            
Reads all the input files (*.<EXTENSION>) in the directory, and applies indicated templates to those files.

Input files are of the form::

    [template]
    
    templatefile = <template filename>
    outputdir = <output directory>
    outputext = <output extension>
    
    [substitutions]
    
    sub1 = <sub1 text>
    
All strings of form {{{sub1}}} in the template file are substituted with <sub1 text> (e.g.).
The format of the template file is defined in http://mustache.github.io/mustache.5.html .  Due to the nature of the
input file, not all mustache features are supported.  E.g., there is currently no way to configure lists.

The input file follows the rules for configuration files defined in http://docs.python.org/2/library/configparser.html#module-ConfigParser .
As an example, long substitutions can be extended over several lines.
After the first line, the subsequent lines of the same substitution must start with some white space.

NOTE: all white space at the beginning of each line is deleted when applied to the template.  This is due to implementation of python's ConfigParser class.
    
'''

# standard
import argparse
import glob
import os.path

# pypi
from configparser import ConfigParser
import pystache

# home grown
from . import version

# github
# other

TEMPLATESEC = 'template'
SUBSEC = 'substitutions'

#----------------------------------------------------------------------
def main():
#----------------------------------------------------------------------
    '''
    add key to key file
    '''

    parser = argparse.ArgumentParser(version='{0} {1}'.format('loutilities',version.__version__))
    #parser = argparse.ArgumentParser()
    parser.add_argument('-e','--extension',help='extension for files to be processed (default %(default)s)',default='htmli')
    args = parser.parse_args()
    
    # act on arguments
    extension = args.extension

    # get list of files with indicated extension
    files = glob.glob('*.{}'.format(extension))
    
    # process each file
    for f in files:
        # file is a config file
        cfg = ConfigParser()
        cfg.read(f)
        
        templatefile = cfg.get(TEMPLATESEC,'templatefile')
        outputdir = cfg.get(TEMPLATESEC,'outputdir')
        outputext = cfg.get(TEMPLATESEC,'outputext')
        
        # determine output path
        thisbase,thisext = os.path.splitext(f)
        outputfile = thisbase + '.' + outputext
        outputpath = os.path.join(outputdir,outputfile)
        
        # each option within SUBSEC represents a substitution from the template file
        substitutions = {}
        for sub in cfg.options(SUBSEC):
            substitutions[sub] = cfg.get(SUBSEC,sub)
            
        # read the template and make the substitutions
        TEMPLATE = open(templatefile)
        template = TEMPLATE.read()
        outstring = pystache.render(template,**substitutions)
        TEMPLATE.close()
        
        # check if the file needs to be updated
        # update is needed if this is a new file, or if it has changed
        updateneeded = True
        if os.path.exists(outputpath):
            CURR = open(outputpath)
            curr = CURR.read()
            CURR.close()
            # if file hasn't change, squelch update
            if curr == outstring:
                updateneeded = False
            
        # update file, if new, or if it's changed
        if updateneeded:
            OUT = open(outputpath,'w')
            OUT.write(outstring)
            OUT.close()

# ###############################################################################
# ###############################################################################
if __name__ == "__main__":
    main()




#!/usr/bin/python
###########################################################################################
# sqlalchemy_helpers  -- helper functions for sqlalchemy access
#
#       Date            Author          Reason
#       ----            ------          ------
#       06/30/18        Lou King        Create (copied from https://github.com/louking/rrwebapp/blob/master/rrwebapp/racedb.py)
#
#   Copyright 2018 Lou King.  All rights reserved
###########################################################################################
'''
sqlalchemy_helpers  -- helper functions for sqlalchemy access
===================================================

'''

# standard

# pypi
from sqlalchemy.orm import object_mapper

# github

# other

# home grown

class dbConsistencyError(Exception): pass
class parameterError(Exception): pass

#----------------------------------------------------------------------
def getunique(session, model, **kwargs):
#----------------------------------------------------------------------
    '''
    retrieve a row from the database, raising exception of more than one row exists for query criteria
    
    :param session: session within which update occurs
    :param model: table model
    :param kwargs: query criteria
    
    :rtype: single instance of the row, or None
    '''

    instances = session.query(model).filter_by(**kwargs).all()

    # error if query returned multiple rows when it was supposed to be unique
    if len(instances) > 1:
        raise dbConsistencyError('found multiple rows in {0} for {1}'.format(model,kwargs))
    
    if len(instances) == 0:
        return None
    
    return instances[0]

#----------------------------------------------------------------------
def update(session, model, oldinstance, newinstance, skipcolumns=[]):
#----------------------------------------------------------------------
    '''
    update an existing element based on kwargs query
    
    :param session: session within which update occurs
    :param model: table model
    :param oldinstance: instance of table model which was found in the db
    :param newinstance: instance of table model with updated fields
    :param skipcolumns: list of column names to update
    :rtype: boolean indicates whether any fields have changed
    '''

    updated = False
    
    # update all columns except those we were told to skip
    for col in object_mapper(newinstance).columns:
        # skip indicated keys
        if col.key in skipcolumns: continue
        
        # if any columns are different, update those columns
        # and return to the caller that it's been updated
        if getattr(oldinstance,col.key) != getattr(newinstance,col.key):
            setattr(oldinstance,col.key,getattr(newinstance,col.key))
            updated = True
    
    return updated

#----------------------------------------------------------------------
def insert_or_update(session, model, newinstance, skipcolumns=[], **kwargs):
#----------------------------------------------------------------------
    '''
    insert a new element or update an existing element based on kwargs query
    
    :param session: session within which update occurs
    :param model: table model
    :param newinstance: instance of table model which is to become representation in the db
    :param skipcolumns: list of column names to skip checking for any changes
    :param kwargs: query criteria
    '''


    # get instance, if it exists
    instance = getunique(session,model,**kwargs)
    
    # remember if we update anything
    updated = False

    # found a matching object, may need to update some of its attributes
    if instance is not None:
        updated = update(session,model,instance,newinstance,skipcolumns)
    
    # new object, just add to database
    else:
        session.add(newinstance)
        updated = True

    if updated:
        session.flush()
        
    return updated



'''
applogging - define logging for the application
================================================
'''
# standard
import logging
from logging.handlers import SMTPHandler
from logging import Formatter
from logging.handlers import TimedRotatingFileHandler
from flask import current_app

# pypi
from flask_security import current_user

# homegrown

class UserFormatter(Formatter):
    # https://flask.palletsprojects.com/en/1.1.x/logging/#injecting-request-information
    def format(self, record):
        if current_user.is_authenticated:
            record.user_email = current_user.email
        else:
            record.user_email = 'anonymous'
        return super().format(record)

# ----------------------------------------------------------------------
def setlogging():
    # ----------------------------------------------------------------------

    # this is needed for any INFO or DEBUG logging
    current_app.logger.setLevel(logging.DEBUG)

    # uncomment next two lines to debug sqlalchemy pool usage
    # logging.basicConfig()
    # logging.getLogger('sqlalchemy.pool').setLevel(logging.DEBUG)

    # # patch werkzeug logging -- not sure why this is being bypassed in werkzeug._internal._log
    # *** for some reason the following code caused debug pin not to be shown, see https://github.com/louking/rrwebapp/issues/300
    # werkzeug_logger = logging.getLogger('werkzeug')
    # werkzeug_logger.setLevel(logging.INFO)

    # set up logging
    ADMINS = current_app.config['EXCEPTION_EMAIL']
    application = current_app.config['APP_LOUTILITY']
    # to test, set if to True, and change run environment to production
    # if True:
    if not current_app.debug:
        if 'MAIL_SERVER' in current_app.config and 'MAIL_PORT' in current_app.config:
            mailhost = (current_app.config['MAIL_SERVER'], current_app.config['MAIL_PORT'])
        elif 'MAIL_SERVER' in current_app.config:
            mailhost = current_app.config['MAIL_SERVER']
        else:
            mailhost = 'localhost'
        credentials = None
        if 'MAIL_USERNAME' in current_app.config and 'MAIL_PASSWORD' in current_app.config:
            credentials = (current_app.config['MAIL_USERNAME'], current_app.config['MAIL_PASSWORD'])
        mail_handler = SMTPHandler(mailhost,
                                   'noreply@steeplechasers.org',
                                   ADMINS, f'{application} exception encountered',
                                   credentials = credentials)
        if 'LOGGING_LEVEL_MAIL' in current_app.config:
            mailloglevel = current_app.config['LOGGING_LEVEL_MAIL']
        else:
            mailloglevel = logging.ERROR
        mail_handler.setLevel(mailloglevel)
        mail_handler.setFormatter(UserFormatter('''
Message type:       %(levelname)s
Location:           %(pathname)s:%(lineno)d
Module:             %(module)s
Function:           %(funcName)s
Time:               %(asctime)s
User:               %(user_email)s

Message:

%(message)s
        '''))
        current_app.logger.addHandler(mail_handler)
        current_app.config['LOGGING_MAIL_HANDLER'] = mail_handler

        logpath = None
        if 'LOGGING_PATH' in current_app.config:
            logpath = current_app.config['LOGGING_PATH']

        if logpath:
            # file rotates every Monday
            file_handler = TimedRotatingFileHandler(logpath, when='W0', delay=True)
            if 'LOGGING_LEVEL_FILE' in current_app.config:
                fileloglevel = current_app.config['LOGGING_LEVEL_FILE']
            else:
                fileloglevel = logging.WARNING
            file_handler.setLevel(fileloglevel)
            current_app.logger.addHandler(file_handler)
            current_app.config['LOGGING_FILE_HANDLER'] = file_handler

            file_handler.setFormatter(UserFormatter(
                '%(asctime)s %(levelname)s: %(user_email)s %(message)s '
                '[in %(pathname)s:%(lineno)d]'
            ))


